@@ -1,11 +1,11 @@
 /*
- * Strobe (c) 1995 Julian Assange (proff@suburbia.net),
+ * Strobe (c) 1995-1997 Julian Assange (proff@suburbia.net),
  * All rights reserved.
  *
  * $ cc strobe.c -o strobe
  */
 
-#define VERSION "1.03"
+#define VERSION "1.05"
 
 #include <stdio.h>
 #include <unistd.h>
@@ -23,16 +23,19 @@
 #include <arpa/inet.h>
 #include <netdb.h>
 #include <string.h>
+#include <strings.h>
 #include <errno.h>
 
-#if defined(solaris) || defined(linux) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__GCC__)
+#if defined(solaris) || defined(linux) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__GCC__) || defined(__GNUC__)
 #  define fvoid void
 #else
 #  define fvoid
 extern int optind;
 extern char *optarg;
 #endif
-#define bool char
+#define bool int
+#define FALSE 0
+#define TRUE  1
 
 #ifndef INADDR_NONE
 #  define INADDR_NONE ((unsigned long)-1)
@@ -55,6 +58,7 @@ extern char *optarg;
 #endif
 
 int a_timeout = 20;
+int a_data_timeout = 30;
 char *a_output = NULL;
 char *a_services = "strobe.services";
 char *a_input = NULL;
@@ -64,7 +68,12 @@ int a_end = 65535;
 int a_sock_max = 64;
 int a_abort = 0;
 int a_bindport = 0;
-char *a_bindaddr= NULL;
+int a_capture = 1024;
+int a_wrap = 79;
+int a_lines = 1;
+char *a_bindaddr = NULL;
+char *a_dircap = NULL;
+char *a_views = "all,hosts,networks,ports";
 struct in_addr bindaddr;
 bool f_linear = 0;
 bool f_verbose = 0;
@@ -75,6 +84,7 @@ bool f_quiet = 0;
 bool f_delete_dupes = 0;
 bool f_minimise = 0;
 bool f_dontgetpeername = 0;
+bool f_hexdump = 0;
 
 int connects = 0;
 int hosts_done = 0;
@@ -82,7 +92,8 @@ int attempts_done = 0;
 int attempts_outstanding = 0;
 struct timeval time_start;
 
-fd_set set_sel;
+fd_set set_sel_check_r;
+fd_set set_sel_check_w;
 fd_set set_sel_r;
 fd_set set_sel_w;
 
@@ -92,6 +103,8 @@ char **Argv;
 
 FILE *fh_input;
 
+char *capture_buf;
+
 #define HO_ACTIVE 1
 #define HO_ABORT 2
 #define HO_COMPLETING 4
@@ -101,6 +114,8 @@ struct hosts_s
     char *name;
     struct in_addr in_addr;
     int port;
+    int *portlist;
+    int portlist_n,portlist_alloc;
     int portlist_ent;
     struct timeval time_used;
     struct timeval time_start;
@@ -111,11 +126,12 @@ struct hosts_s
     time_t notice_abort;
     int status;
 };
-struct hosts_s ho_initial;
+struct hosts_s ho_initial; /* the null template */
 struct hosts_s *hosts;
 
 #define HT_SOCKET 1
 #define HT_CONNECTING 2
+#define HT_CONNECTED 4
 
 struct htuple_s
 {
@@ -127,6 +143,12 @@ struct htuple_s
     struct timeval sock_start;
     int timeout;
     struct hosts_s *host;
+    int data_lines;
+    int data_len;
+    u_char *data;
+    int send_pos;
+    int send_len;
+    u_char *send;
 };
 
 struct htuple_s ht_initial;
@@ -146,6 +168,8 @@ struct port_desc_s **port_descs;
 int *portlist = NULL;
 int portlist_n = 0;
 
+void display_port_sw (struct htuple_s *h);
+
 char *
 Srealloc (ptr, len)
   char *ptr;
@@ -193,34 +217,26 @@ sock_unblock (sfd)
 }
 
 int
-timeval_subtract (result, x, y) /* from gnu c-lib info.texi */
+timeval_subtract (result, x, y) /* why not floating point?  */
   struct timeval *result, *x, *y;
 {
-/* Perform the carry for the later subtraction by updating y. */
-if (x->tv_usec < y->tv_usec) {
- int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
- y->tv_usec -= 1000000 * nsec;
- y->tv_sec += nsec;
-}
-if (x->tv_usec - y->tv_usec > 1000000) {
- int nsec = (y->tv_usec - x->tv_usec) / 1000000;
- y->tv_usec += 1000000 * nsec;
- y->tv_sec -= nsec;
-}
-
-/* Compute the time remaining to wait.
-  `tv_usec' is certainly positive. */
-result->tv_sec = x->tv_sec - y->tv_sec;
-result->tv_usec = x->tv_usec - y->tv_usec;
-
+  result->tv_usec = x->tv_usec - y->tv_usec;
+  result->tv_sec = x->tv_sec - y->tv_sec;
+  if (result->tv_usec<0)
+  {
+    result->tv_usec+=1000000;
+    result->tv_sec --;
+  }
 /* Return 1 if result is negative. */
-return x->tv_sec < y->tv_sec;
+return result->tv_sec < 0;
 }
 
 fvoid
 attempt_clear (h)
   struct htuple_s *h;
 {
+    if (h->status & HT_CONNECTED)
+        display_port_sw (h);
     if (h->status & HT_SOCKET)
     {
 	struct timeval tv1, tv2;
@@ -239,12 +255,15 @@ attempt_clear (h)
 	sock_unblock (h->sfd);
 /*	shutdown (h->sfd, 2); */
 	close (h->sfd);
-        if (FD_ISSET(h->sfd, &set_sel))
+        if ( (FD_ISSET(h->sfd, &set_sel_check_r)) || (FD_ISSET(h->sfd, &set_sel_check_w)) )
 	{
-	     FD_CLR (h->sfd, &set_sel);
+	     FD_CLR (h->sfd, &set_sel_check_r);
+             FD_CLR (h->sfd, &set_sel_check_w);
 	     attempts_outstanding--;
 	}
     }
+    if (h->data)
+        free(h->data);
     *h = ht_initial;
 }
 
@@ -277,7 +296,8 @@ fdsets_init ()
 {
     FD_ZERO(&set_sel_r); /* yes, we have to do this, despite the later */
     FD_ZERO(&set_sel_w); /* assisgnments */
-    FD_ZERO(&set_sel);
+    FD_ZERO(&set_sel_check_r);
+    FD_ZERO(&set_sel_check_w);
 }
 
 int
@@ -311,10 +331,10 @@ sc_connect (h)
 			exit(1);
 		return 0;
 	}
-    setsockopt (h->sfd, SOL_SOCKET, SO_REUSEADDR, (char *) &sopts1, sizeof (sopts1));
     sa_in.sin_addr = h->in_addr;
     sa_in.sin_port = htons (h->port);
 
+    h->host->attempts++;
     if (connect (h->sfd, (struct sockaddr *) &sa_in, sizeof (sa_in)) == -1)
     {
 	switch (errno)
@@ -325,15 +345,13 @@ sc_connect (h)
 	case ETIMEDOUT:
 	case ECONNREFUSED:
 	case EADDRNOTAVAIL:
-	case EPERM:
 	    if (f_verbose)
 	    {
 		fprintf(stderr, "%s:%d ", h->name, h->port);
 		perror("");
 	    }
-	    h->host->attempts++;
 	    attempt_clear (h);
-	    return 1;
+	    return 0;
 	default:
 	    if (!f_quiet)
 	    {
@@ -344,97 +362,521 @@ sc_connect (h)
 	    return 0;
 	}
     }
-    h->host->attempts++;
     h->status |= HT_CONNECTING;
     sock_block (h->sfd);
-    FD_SET(h->sfd, &set_sel);
+    FD_SET(h->sfd, &set_sel_check_r);
+    FD_SET(h->sfd, &set_sel_check_w);
     attempts_outstanding++;
     return 1;
 }
 
+void
+gen_port_simple (h, buf)
+  struct htuple_s *h;
+  char *buf;
+{
+    sprintf (buf, "%.128s %5d", h->name, h->port);
+}
+
+int
+display_port_simple (h, fh)
+  struct htuple_s *h;
+  FILE *fh;
+{
+    char buf[256];
+    gen_port_simple(h, buf);
+    fputs(buf, fh);
+    return strlen(buf);
+}
+
+void
+display_port (h, fh)
+  struct htuple_s *h;
+  FILE *fh;
+{
+    int ds_len;
+    ds_len = display_port_simple(h, fh);
+    if (f_minimise)
+	fputc('\n', fh);
+    else
+    {
+        struct port_desc_s *pd;
+    	if ((pd = port_descs[h->port]))
+        {
+    	    fprintf (fh, " %-12s %s\n", pd->portname, pd->name);
+	    while (!f_delete_dupes && (pd=pd->next))
+	        fprintf (fh, "%*s %-12s %s\n", ds_len, "", pd->portname, pd->name);
+    	}
+    	else
+    	    fprintf (fh, " %-12s %s\n", "unassigned", "unknown");
+    }
+}
+
+u_char *
+conv_char(c, c2)
+u_char c;
+u_char c2;
+{
+        static u_char b[8];
+	bzero(b, sizeof b);
+	if (c=='\\')
+	{
+		b[0] = c;
+		b[1] = c;
+	} else
+	if (c>=32 && c<127)
+	{
+		b[0] = c;
+	} else
+	if (c == '\r')
+	{
+		b[0] = '\\';
+		b[1] = 'r';
+	} else
+	if (c == '\n')
+	{
+		b[0] = '\\';
+		b[1] = 'n';
+	} else
+	if (c == '\t')
+	{
+		b[0] = '\\';
+		b[1] = 't';
+	} else
+	if (c == 8)
+	{
+		b[0] = '\\';
+		b[1] = 'b';
+	} else
+	if (c<32 || c >= 127)
+	{
+		/* remove possible ambiguity \n or \nn then digit */
+		if (c<100 && isdigit(c2))
+			sprintf (b, "\\%03d", (int)c);
+		else
+			sprintf (b, "\\%d", (int)c);
+	} else
+	{
+		fprintf (stderr, "internal error in print_char");
+	}
+	return b;
+}
+
+void
+asciidump (h, fh)
+  struct htuple_s *h;
+  FILE *fh;
+{
+    int n;
+    int col = 0;
+    char buf[256];
+    int buf_len;
+    char *p=h->data;
+    gen_port_simple(h, buf);
+    buf_len = strlen(buf);
+    for (n=0; n<h->data_len; n++) {
+        if (col == 0)
+	{
+	    if (!a_dircap)
+	    {
+	        if (f_minimise)
+		{
+		    fputs(buf, fh);
+		    col = buf_len;
+	        }
+		else
+		{
+		    fprintf(fh, "%*s->", buf_len+1, "");
+		}
+		fputc(' ', fh);
+	    }
+	}
+    	fputs(conv_char(p[n], (n==h->data_len-1)? '\0': p[n+1]), fh);
+	col++;
+	if (a_wrap)
+	{
+	    if (col >= a_wrap)
+	    {
+	        fputc('\n', fh);
+		col = 0;
+	    }
+	    else
+	    {
+		switch (p[n])
+		{
+		 case '\0':
+		 case '\n':
+		    fputc('\n', fh);
+		    col = 0;
+		}
+            }
+       }
+    }
+   if (col !=0)
+       fputc('\n', fh);
+}
+
+void
+hexdump(h, fh)
+  struct htuple_s *h;
+  FILE *fh;
+{
+	int n;
+    char buf[256];
+    int buf_len;
+    gen_port_simple(h, buf);
+    buf_len = strlen(buf);
+	for (n = 0; n < h->data_len; n+=12)
+	{
+		int y;
+	    if (!a_dircap)
+	    {
+		if (f_minimise)
+		{
+		    fputs(buf, fh);
+		}
+		else
+		{
+		    fprintf(fh, "%*s->", buf_len+1, "");
+		}
+		fputc(' ', fh);
+	    }
+		for (y=0; y<12; y++)
+		{
+			if (y+n< h->data_len)
+				fprintf(fh, "%02X ", h->data[n+y]);
+			else
+				fputs("   ", fh);
+			if (y%4==3 && y!=11)
+				fputc(' ', fh);
+		}
+		fputs(": ", fh);
+		for (y=0; y<12; y++)
+		{
+			if (y+n< h->data_len)
+				fprintf(fh, "%c", isprint(h->data[n+y])? h->data[n+y]: '.');
+			else
+				fputc(' ', fh);
+		}
+		fputc('\n', fh);
+	}
+	fputc('\n', fh);
+}
+
+char *
+trslash (s)
+  char *s;
+{
+    char *p=s;
+    for (;*s; s++)
+        if (*s == '/')
+	    *s = ',';
+    return p;
+}
+	    
+/*
+ * recursively build directory hierarchy, starting at leaf
+ */
+
+bool
+blddir (name)
+  char *name;
+{
+	char *p;
+
+	if ((p = strrchr (name, '/')) == NULL)
+		return FALSE;
+	*p = '\0';
+	if (mkdir (name, (mode_t) 0775) == -1)
+		if (!blddir (name))
+		{
+			*p = '/';
+			return FALSE;
+		}
+	if (mkdir (name, (mode_t) 0775) == -1)
+	{
+		if (errno != EEXIST)
+		{
+			fprintf(stderr, "error building directory hierarchy %s", name);
+			*p = '/';
+			return FALSE;
+		}
+	}
+	*p = '/';
+	return TRUE;
+}
+
+bool
+makeln (char *from, char *to)
+{
+    if (link(from, to) !=0)
+    {
+        blddir(to);
+	if (link(from, to) !=0)
+	{
+	    perror(to);
+	    return FALSE;
+	}
+    }
+    return TRUE;
+}
+
+void
+display_port_sw (h)
+  struct htuple_s *h;
+{
+    char master[1024];
+    FILE *fh = stdout;
+
+    if (a_dircap)
+    {
+        char buf[2048];
+	char in[64];
+	char *p;
+        struct port_desc_s *pd;
+	sprintf(master, "%.512s/strobe.%d", a_dircap, (int)getpid());
+	fh = fopen(master, "w");
+	if (!fh)
+	{
+	    blddir(master);
+	    fh = fopen(master, "w");
+	    if (!fh)
+	    {
+	        perror(master);
+		return;
+	    }
+	}
+	strcpy(in, inet_ntoa(h->host->in_addr));
+	for (p=in; *p; p++)
+	    if (*p == '.')
+	        *p = '/';
+    	pd = port_descs[h->port];
+	if (strstr(a_views, "networks"))
+	{
+	        if (f_minimise)
+			sprintf(buf, "%.512s/networks/%s/%s-%.256s/%05d", a_dircap, in, inet_ntoa(h->host->in_addr),
+			        h->host->name, h->port);
+		else
+			sprintf(buf, "%.512s/networks/%s/%s-%.256s/%05d-%.200s-%.200s", a_dircap, in, inet_ntoa(h->host->in_addr),
+			        h->host->name, h->port, pd? port_descs[h->port]->portname: "unknown", pd? trslash(port_descs[h->port]->name): "");
+		makeln(master, buf);
+	}
+	if (strstr(a_views, "hosts"))
+	{
+	        if (f_minimise)
+			sprintf(buf, "%.512s/hosts/%s-%.256s/%05d", a_dircap, inet_ntoa(h->host->in_addr),
+			        h->host->name, h->port);
+		else
+			sprintf(buf, "%.512s/hosts/%s-%.256s/%05d-%.200s-%.200s", a_dircap, inet_ntoa(h->host->in_addr),
+			        h->host->name, h->port, pd? port_descs[h->port]->portname: "unknown", pd? trslash(port_descs[h->port]->name): "");
+		makeln(master, buf);
+	}
+	if (strstr(a_views, "all"))
+	{
+	        if (f_minimise)
+			sprintf(buf, "%.512s/all/%.64s-%.200s-%05d", a_dircap, inet_ntoa(h->host->in_addr),
+				h->host->name, h->port);
+		else
+			sprintf(buf, "%.512s/all/%.64s-%.200s-%05d-%.200s-%.200s", a_dircap, inet_ntoa(h->host->in_addr),
+				h->host->name, h->port, pd? port_descs[h->port]->portname: "unknown", pd? trslash(port_descs[h->port]->name): "");
+		makeln(master, buf);
+	}
+	if (strstr(a_views, "ports"))
+	{
+	        if (f_minimise)
+			sprintf(buf, "%.512s/ports/%05d/%s-%.200s", a_dircap, h->port, inet_ntoa(h->host->in_addr),
+				h->host->name);
+		else
+			sprintf(buf, "%.512s/ports/%05d-%.200s-%.200s/%.64s-%.200s", a_dircap, h->port, 
+				pd? port_descs[h->port]->portname: "unknown", pd? trslash(port_descs[h->port]->name): "",
+				inet_ntoa(h->host->in_addr), h->host->name);
+		makeln(master, buf);
+	}
+    }
+    if (h->status&HT_CONNECTED)
+    {
+	    if (!f_minimise && !a_dircap)
+		    display_port(h, fh);
+	    if (h->data_len>0)
+	    {
+		    if (f_hexdump)
+			hexdump(h, fh);
+		    else
+			asciidump(h, fh);
+	    }
+    } 
+    if (a_dircap)
+    {
+        fclose(fh);
+        unlink(master);
+    }
+    if (!(h->status&HT_CONNECTED))
+        display_port(h,fh);
+}
+
+void
+init_capture_tcp (h)
+  struct htuple_s *h;
+{
+    h->status&=~HT_CONNECTING;
+    h->status|=HT_CONNECTED;
+    FD_CLR (h->sfd, &set_sel_check_w);
+}
+
+void
+capture_tcp (h)
+  struct htuple_s *h;
+{
+    int cc;
+    bool f_drop = 0;
+    cc = recv(h->sfd, capture_buf, a_capture, 0);
+    if (cc == 0)
+    {
+    	attempt_clear (h);
+	return;
+    }
+    if (cc < 0)
+    {
+    	switch (errno)
+	{
+	 case EAGAIN:
+	 case EINTR:
+	    break;
+	 default:
+	    attempt_clear (h);
+	}
+	return;
+    }
+    if (h->data_len+cc>a_capture)
+    {
+        cc = a_capture - h->data_len;
+	f_drop = 1;
+    }
+    if (a_lines)
+    {
+        int n;
+	for (n=0; n<cc; n++)
+	    if (capture_buf[n] == '\n' ||
+	        capture_buf[n] == '\0')
+	        h->data_lines++;
+    }
+    h->data_len+=cc;
+    if (h->data)
+        h->data = Srealloc(h->data, h->data_len);
+    else
+        h->data = Smalloc(h->data_len);
+    memcpy(h->data+h->data_len-cc, capture_buf, cc);
+    if (f_drop || h->data_lines >= a_lines)
+        attempt_clear (h);
+}
+
 int
 gatherer_tcp (h)
   struct htuple_s *h;
 {
-    struct port_desc_s *pd;
-    if (f_minimise)
-	printf ("%s\t%d\n", h->name, h->port);
-    else
-    {
-    	if ((pd = port_descs[h->port]))
-        {
-    	    	printf ("%-30s %-16s %5d/tcp %s\n", h->name, pd->portname, h->port, pd->name);
-	    while (!f_delete_dupes && !f_minimise && (pd=pd->next))
-	        printf ("#%-29s %-16s %5d/tcp %s\n", h->name, pd->portname, h->port, pd->name);
-    	}
-    	else
-	    printf ("%-30s %-16s %5d/tcp unassigned\n", h->name, "unknown", h->port);
-    }
     h->host->connects++;
     connects++;
-    attempt_clear (h);
+    if (a_capture)
+    	init_capture_tcp (h);
+    else 
+    {
+        display_port_sw (h);
+	attempt_clear (h);
+    }
     return 1;
 }
 
-bool
-gather ()
+int
+gather (timeout_secs)
+  int timeout_secs;
 {
     struct timeval timeout;
     struct htuple_s *h;
     int n;
+    int last = -1;
     int selected;
     time_t tim;
 
     if (!attempts_outstanding) return 1;
-    set_sel_r=set_sel_w=set_sel;
-    timeout.tv_sec = 0;
-    timeout.tv_usec = 250000; /* 1/4 of a second */
+    set_sel_r=set_sel_check_r;
+    set_sel_w=set_sel_check_w;
 
+    if (timeout_secs)
+    {
+        timeout.tv_sec = timeout_secs;
+        timeout.tv_usec = 0;
+    }
+    else
+    {
+        timeout.tv_sec = 0;
+        timeout.tv_usec = 250000; /* 1/4 of a second */
+    }
+    
+    
     selected = select (FD_SETSIZE, &set_sel_r, &set_sel_w, 0, &timeout);
-    /* Look for timeouts */
+    if (selected<0)
+	perror ("select");
 
     tim = time (NULL);
     for ( n = 0 ; n < a_sock_max; n++ )
     {
-      h = &attempt[n];
-      if ((h->status & HT_SOCKET) &&
-	  ((h->sock_start.tv_sec + h->timeout) < tim))
- 	attempt_clear (h);
-    }
-
-    switch (selected)
-    {
-    case -1:
-	perror ("select");
-	return 0;
-    case 0:
-	return 1;
-    }
-    for (n = 0; selected && (n < a_sock_max); n++)
-    {
-	h = &attempt[n];
-	if (h->status & HT_CONNECTING)
+        h = &attempt[n];
+        if (h->status & HT_CONNECTED)
+        {
+            if (!FD_ISSET(h->sfd,&set_sel_r))
+            {
+              if ( (tim - h->sock_start.tv_sec) >= h->timeout)
+              {
+                attempt_clear(h);
+              }
+              continue;
+            }
+        }
+	if (selected>0 && h->status & (HT_CONNECTING|HT_CONNECTED))
 	{
-	    if (FD_ISSET (h->sfd, &set_sel_r) || FD_ISSET (h->sfd, &set_sel_w))
+	    if (FD_ISSET (h->sfd, &set_sel_r) ||
+                FD_ISSET (h->sfd, &set_sel_w))
 	    {
 		struct sockaddr_in in;
 		int len = sizeof (in);
 		selected--;
-		/* select() lies occasionaly
+
+                    
+                /* select() lies occasionaly
                  */
 		if (!f_dontgetpeername) /* but solaris2.3 crashes occasionally ;-| */
 		{
 			if (getpeername (h->sfd, (struct sockaddr *) &in, &len) == 0)
-		    	    gatherer_tcp (h);
+			{
+			    if ((h->status & HT_CONNECTED) && FD_ISSET (h->sfd, &set_sel_r))
+                                capture_tcp (h);
+			    else
+		    	        gatherer_tcp (h);
+		        }
 			else
 		    	    attempt_clear (h);
 		}
 		else
-		    	gatherer_tcp (h);
+		{
+		    if ((h->status & HT_CONNECTED) &&
+			FD_ISSET (h->sfd, &set_sel_r))
+			capture_tcp (h);
+		    else
+			gatherer_tcp (h);
+		}
+	    }
+	    last = n;
+	} else
+	{
+	    if ((h->status & HT_SOCKET) &&
+	        ((h->sock_start.tv_sec + h->timeout) < tim))
+	    {
+	        attempt_clear (h);
+		last = n;
 	    }
 	}
     }
-    return 1;
+    return last;
 }
 
 bool
@@ -443,33 +885,33 @@ add_attempt (add)
 {
     struct htuple_s *h;
     static time_t oldtime;
-    static int n;
+    int ret;
     for (;;)
     {
-	for (; n < a_sock_max; n++)
+ 	int n;
+        int last;
+	for (n=0; n < a_sock_max; n++)
 	{
 	    h = &attempt[n];
 	    if (!h->status)
 		goto foundfree;
 	}
-	n = 0;
-	gather ();
-	continue;
-      foundfree:
-	*h = *add;
-	if (!sc_connect (h))
-	{
-	    gather ();
-	    continue;
+	last = gather (a_timeout);
+	oldtime = time(NULL);
+	if (last != -1) {
+	   h = &attempt[last];
+	   goto foundfree;
 	}
-	if ((oldtime + 1) < time (NULL))
-	{
-	    oldtime = time (NULL);
-	    gather ();
-	}
-	break;
     }
-    return 1;
+    foundfree:
+    *h = *add;
+    ret = sc_connect (h);
+    if (oldtime+1<time(NULL)) {
+	if (oldtime > 0)
+		gather (0);
+	oldtime = time(NULL);
+    }
+    return ret;
 }
 
 int
@@ -498,7 +940,7 @@ wait_end (t)
     st = time (NULL);
     while ((st + t) > time (NULL))
     {
-	gather ();
+	gather (a_timeout);
 	if (attempts_outstanding<1) break;
     }
 }
@@ -528,7 +970,6 @@ char *
 next_host ()
 {
     static char lbuf[512];
-    hosts_done++;
     if (a_input)
     {
 	int n;
@@ -543,14 +984,53 @@ reread:
 	n = strcspn (lbuf, " \t\n\r");
 	if (n)
 	    lbuf[n] = '\0';
+        hosts_done++;
 	return lbuf;
     }
     if ( host_n >= Argc )
       return NULL;
 
+    hosts_done++;
     return Argv[host_n++];
 }
 
+int
+next_port (h)
+struct hosts_s *h;
+{
+    int n;
+    if (f_fast)
+        return (++h->portlist_ent>portlist_n)?-1:portlist[h->portlist_ent-1];
+    else if (h->portlist)
+    {
+        return (++h->portlist_ent>h->portlist_n)?-1:h->portlist[h->portlist_ent-1];
+    }
+    else
+    {
+        for (n = h->port; ++n <= a_end;)
+        {
+            return n;
+        }
+    }
+    return -1;
+}
+
+int
+add_port(h,p)
+  struct hosts_s *h;
+  int p;
+{
+    if (!p)
+        return 0;
+    if (h->portlist_n == h->portlist_alloc)
+    {
+        h->portlist_alloc += 20;
+        h->portlist=(int *)Srealloc(h->portlist,h->portlist_alloc*sizeof(int));
+    }
+    h->portlist[h->portlist_n++]=p;
+    return p;
+}
+
 bool
 host_init (h, name, nocheck)
   struct hosts_s *h;
@@ -558,7 +1038,61 @@ host_init (h, name, nocheck)
   bool nocheck;
 {
     int n;
+    char *ports;
+    
     *h=ho_initial;
+    if ((ports=strchr(name,':')))
+    {
+        char *pstart, *minus;
+        int i, stopnow=0;
+        int lastport;
+        
+        *ports=0;
+        pstart=++ports;
+        while(!stopnow)
+        {
+            switch(*ports)
+            {
+                case '\0':
+                    stopnow=1;
+                    /* FALL THROUGH */
+                case ',':
+                case ':':
+                    *ports=0;
+                    if ((minus=strchr(pstart,'-')))
+                    {
+                        *minus=0;
+                        lastport=atoi(minus++);
+
+                        for(i=atoi(pstart);i<=lastport;i++)
+                            if (!add_port(h,i))
+                            {
+                                stopnow=2;
+                                break;
+                            }
+                    }
+                    else
+                    {
+                        if (!add_port(h,atoi(pstart)))
+                            stopnow=2;
+                    }
+                    pstart=ports+1;
+                    break;
+                case '0': case '1': case '2': case '3': case '4':
+                case '5': case '6': case '7': case '8': case '9':
+                case '-':
+                    break;
+                default:
+                    stopnow=2;
+            }
+            ports++;
+        }
+        if (stopnow==2)
+        {
+            fprintf(stderr,"Couldn't parse port specifier\n");
+            exit(1);
+        }
+    }
     h->in_addr = resolve (name);
     if (h->in_addr.s_addr == INADDR_NONE)
 	return 0;
@@ -575,7 +1109,10 @@ host_init (h, name, nocheck)
         }
     h->name = (char *) Smalloc (strlen (name) + 1);
     strcpy (h->name, name);
-    h->port = a_start;
+    if(f_fast || h->portlist)
+        h->port=next_port(h);
+    else
+        h->port = a_start;
     h->status = HO_ACTIVE;
     gettimeofday(&(h->time_start), NULL);
     return 1;
@@ -641,20 +1178,6 @@ bool skip_host(h)
     return 0;
 }
 
-int
-next_port (h)
-struct hosts_s *h;
-{
-    int n;
-    for (n = h->port; ++n <= a_end;)
-    {
-    	if (!f_fast) return n;
-	if (++h->portlist_ent>portlist_n) return -1;
-        return (portlist[h->portlist_ent-1]);
-    }
-    return -1;
-}
-
 fvoid
 scan_ports_linear ()
 {
@@ -699,12 +1222,13 @@ scan_ports_paralell ()
     struct timeval smallest_val;
     int smallest_cnt;
     char *name;
-    struct hosts_s *h, *smallest = &hosts[0];
-    while (smallest)
+    struct hosts_s *h, *smallest;
+    struct hosts_s *anyhost;
+    do 
     {
 	smallest_val.tv_sec=0xfffffff;
 	smallest_val.tv_usec=0;
-	for (n = 0, smallest_cnt = 0xfffffff, smallest = NULL; n < a_sock_max; n++)
+	for (n = 0, smallest_cnt = 0xfffffff, anyhost= smallest = NULL; n < a_sock_max; n++)
 	{
 	    h = &hosts[n];
 
@@ -725,12 +1249,14 @@ scan_ports_paralell ()
 
 	    if (h->name)
 	    {
-		if (((h->time_used.tv_sec < smallest_val.tv_sec) ||
+                anyhost=h;
+		if ((((h->time_used.tv_sec < smallest_val.tv_sec) ||
 		     ((h->time_used.tv_sec == smallest_val.tv_sec) &&
 		      (h->time_used.tv_usec <= smallest_val.tv_usec))) &&
 		    (((h->time_used.tv_sec != smallest_val.tv_sec) &&
 		      (h->time_used.tv_usec != smallest_val.tv_usec)) ||
-		     (h->attempts < smallest_cnt)))
+		     (h->attempts < smallest_cnt)))&&
+                     !(h->status&HO_COMPLETING))
 	        {
 	  	    smallest_cnt = h->attempts;
 		    smallest_val = h->time_used;
@@ -741,16 +1267,15 @@ scan_ports_paralell ()
 
 	if (smallest)
 	{
-	    if (!(smallest->status & HO_COMPLETING))
-	    {
+/* scatter adds connection or calls gather() until one is freed   */
 		scatter (smallest, a_timeout);
 		if ((smallest->port=next_port(smallest))==-1)
 	            smallest->status|=HO_COMPLETING;
-	    }
-	    else
-		gather();
 	}
+        else
+            gather(a_timeout);
     }
+    while(anyhost);
 }
 
 fvoid
@@ -789,29 +1314,24 @@ loaddescs ()
 	    portlist = (int *)Srealloc((char *)portlist, ++portlist_n*sizeof(int));
 	    portlist[portlist_n-1]=port;
 	}
-	if (!f_minimise)
+	pdp = (struct port_desc_s *) Smalloc (sizeof (*pd) + strlen (desc) + 1 + strlen (portname) + 1);
+	if (pd)
 	{
-	    pdp = (struct port_desc_s *) Smalloc (sizeof (*pd) + strlen (desc) + 1 + strlen (portname) + 1);
-	    if (pd)
-	    {
-	        for (; pd->next; pd = pd->next);
-	        pd->next = pdp;
-	        pd = pd->next;
-	    } else 
-	    {
-	        pd = pdp;
-	        port_descs[port] = pd;
-	    } 
-	    pd->next = NULL;
-	    pd->name = (char *) (pd) + sizeof (*pd);
-	    pd->portname = pd->name + strlen(desc)+1;
-	    strcpy (pd->name, desc);
-	    strcpy (pd->portname, portname);
-	} else
-	    port_descs[port] = (struct port_desc_s *)-1;
+	    for (; pd->next; pd = pd->next);
+	    pd->next = pdp;
+	    pd = pd->next;
+	} else 
+	{
+	    pd = pdp;
+	    port_descs[port] = pd;
+	} 
+	pd->next = NULL;
+	pd->name = (char *) (pd) + sizeof (*pd);
+	pd->portname = pd->name + strlen(desc)+1;
+	strcpy (pd->name, desc);
+	strcpy (pd->portname, portname);
     }
-    if (f_minimise) 
-        free (port_descs);
+    fclose (fh);
 }
 
 fvoid
@@ -839,6 +1359,12 @@ usage: %8s [options]\n\
 \t\t[-l(inear)]\n\
 \t\t[-f(ast)]\n\
 \t\t[-a abort_after_port_n]\n\
+\t\t[-c capture_n]\n\
+\t\t[-w wrap_col_n]\n\
+\t\t[-x(heXdump)]\n\
+\t\t[-L capture_lines_n]\n\
+\t\t[-D capture_directory]\n\
+\t\t[-T capture_timeout_n]\n\
 \t\t[-M(ail_author)]\n\
 \t\t[host1 [...host_n]]\n", Argv[0]);
     exit (1);
@@ -852,7 +1378,7 @@ main (argc, argv)
     Argc = argc;
     Argv = argv;
 
-    while ((c = getopt (argc, argv, "o:dvVmgb:e:p:P:a:A:t:n:S:i:lfsqM")) != -1)
+    while ((c = getopt (argc, argv, "o:dvVmgb:e:p:P:a:A:t:n:S:i:lfsqMc:w:xL:D:T:")) != -1)
 	switch (c)
 	{
 	case 'o':
@@ -925,6 +1451,24 @@ main (argc, argv)
 	    fprintf(stderr, "Enter mail to author below. End with ^D or .\n");
 	    system("mail strobe@suburbia.net");
 	    break;
+	case 'c':
+	    a_capture = atoi(optarg);
+	    break;
+	case 'w':
+	    a_wrap = atoi(optarg);
+	    break;
+	case 'x':
+	    f_hexdump = 1;
+	    break;
+	case 'L':
+	    a_lines = atoi(optarg);
+	    break;
+        case 'D':
+            a_dircap = optarg;
+            break;
+	case 'T':
+	    a_data_timeout = atoi (optarg);
+	    break;
 	case '?':
 	default:
 	    fprintf (stderr, "unknown option %s\n", argv[optind-1]);
@@ -934,7 +1478,7 @@ main (argc, argv)
     host_n = optind;
 
     if (!f_quiet)
-        fprintf (stderr, "strobe %s (c) 1995 Julian Assange (proff@suburbia.net).\n", VERSION);
+        fprintf (stderr, "strobe %s (c) 1995-1999 Julian Assange <proff@iq.org>.\n", VERSION);
     if (a_input)
     {
         if ( ! strcmp("-",a_input) ) { /* Use stdin as input file */
@@ -973,6 +1517,8 @@ main (argc, argv)
 	}
 	dup2(fd, 1);
     }
+    if (a_capture)
+	    capture_buf = Smalloc(a_capture);
     attempt = (struct htuple_s *) Smalloc (a_sock_max * sizeof (struct htuple_s));
     attempt_init();
     if (!f_linear)
@@ -980,8 +1526,7 @@ main (argc, argv)
     	hosts = (struct hosts_s *) Smalloc (a_sock_max * sizeof (struct hosts_s));
     	hosts_init();
     }
-    if (!f_minimise || f_fast)
-    	loaddescs ();
+    loaddescs ();
     fdsets_init();
     gettimeofday(&time_start, NULL);
     f_linear ? scan_ports_linear ():
